# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Zoom line search algorithm."""

import dataclasses
import functools
from typing import Any
from typing import Callable
from typing import NamedTuple
from typing import Optional
from typing import Union

import jax
from jax import lax
import jax.numpy as jnp
from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.base import _make_funs_with_aux
from nemos.third_party.jaxopt.jaxopt._src.cond import cond
from nemos.third_party.jaxopt.jaxopt._src.tree_util import tree_single_dtype
from nemos.third_party.jaxopt.jaxopt._src.tree_util import get_real_dtype
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_add_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_scalar_mul
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_vdot_real
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_conj

# pylint: disable=g-bare-generic
# pylint: disable=invalid-name

# Flags to print errors, used in tests
WARNING_PREAMBLE = \
  "WARNING: jaxopt.ZoomLineSearch: "
FLAG_NAN_INF_VALUES = WARNING_PREAMBLE + \
  "NaN or Inf values encountered in function values."
FLAG_INTERVAL_NOT_FOUND = WARNING_PREAMBLE + \
  "No interval satisfying curvature condition." + \
  "Consider increasing maximal possible stepsize of the linesearch."
FLAG_INTERVAL_TOO_SMALL = WARNING_PREAMBLE + \
  "Length of searched interval has been reduced below threshold."
FLAG_CURVATURE_COND_NOT_SATSIFIED = WARNING_PREAMBLE + \
  "Returning stepsize with sufficient decrease " + \
  "but curvature condition not satisfied."
FLAG_NO_STEPSIZE_FOUND = WARNING_PREAMBLE + \
  "Linesearch failed, no stepsize satisfying sufficient decrease found."
FLAG_NOT_A_DESCENT_DIRECTION = WARNING_PREAMBLE + \
  "The linesearch failed because the provided direction " + \
  "is not a descent direction. "

_dot = functools.partial(jnp.dot, precision=lax.Precision.HIGHEST)


def _cubicmin(a, fa, fpa, b, fb, c, fc):
  """Cubic interpolation.

  Finds a critical point of a cubic polynomial
  p(x) = A *(x-a)^3 + B*(x-a)^2 + C*(x-a) + D, that goes through
  the points (a,fa), (b,fb), and (c,fc) with derivative at a of fpa.
  May return NaN (if radical<0), in that case, the point will be ignored.
  Taken from scipy.optimize._linesearch.py.

  Args:
    a: scalar
    fa: value of a function f at a
    fpa: slope of a function f at a
    b: scalar
    fb: value of a function f at b
    c: scalar
    fc: value of a function f at c

  Returns:
    xmin: point at which p'(xmin) = 0
  """
  C = fpa
  db = b - a
  dc = c - a
  denom = (db * dc) ** 2 * (db - dc)
  d1 = jnp.array([[dc**2, -(db**2)], [-(dc**3), db**3]])
  A, B = _dot(d1, jnp.array([fb - fa - C * db, fc - fa - C * dc])) / denom

  radical = B * B - 3.0 * A * C
  xmin = a + (-B + jnp.sqrt(radical)) / (3.0 * A)

  return xmin


def _quadmin(a, fa, fpa, b, fb):
  """Quadratic interpolation.

  Finds a critical point of a quadratic polynomial
  p(x) = B*(x-a)^2 + C*(x-a) + D, that goes through
  the points (a,fa), (b,fb) with derivative at a of fpa.
  Taken from scipy.optimize._linesearch.py.

  Args:
    a: scalar
    fa: value of a function f at a
    fpa: slope of a function f at a
    b: scalar
    fb: value of a function f at b

  Returns:
    xmin: point at which p'(xmin) = 0
  """
  D = fa
  C = fpa
  db = b - a
  B = (fb - D - C * db) / (db**2)
  xmin = a - C / (2.0 * B)
  return xmin


def _set_values(cond, candidate, default):
  def _set_val(x, y):
    return jnp.where(cond, x, y)

  return jax.tree_util.tree_map(_set_val, candidate, default)


def _cond_print(condition, message, **kwargs):
  jax.lax.cond(condition, lambda _: jax.debug.print(message, **kwargs), lambda _: None, None)


class ZoomLineSearchState(NamedTuple):
  """Named tuple containing state information for core loop."""

  iter_num: int
  params: Any  # either initial or final
  value: float  # either initial or final
  grad: Any  # either initial or final

  # unchanged after initialization
  value_init: float  #  (redundant with value, left for readability)
  slope_init: float
  descent_direction: Any

  num_fun_eval: int
  num_grad_eval: int

  decrease_error: float
  curvature_error: float
  error: float
  done: bool
  failed: bool  # comply to semantic used by other line searches

  # Used only during the interval search
  interval_found: bool
  prev_stepsize: float
  prev_value_step: float
  prev_slope_step: float

  # Set up after interval search done, modified during zoom
  low: float
  value_low: float
  slope_low: float
  high: float
  value_high: float
  slope_high: float
  cubic_ref: float
  value_cubic_ref: float

  # Safeguard point: we may not be able to satisfy the curvature condition
  # but we can still return a point that satisfies the decrease condition
  safe_stepsize: float

  aux: Optional[Any] = None  # either initial or final


@dataclasses.dataclass(eq=False)
class ZoomLineSearch(base.IterativeLineSearch):
  """Inexact line search that satisfies sufficient decrease (Armijo) and small curvature (strong Wolfe) conditions.

  Algorithms 3.5, 3.6 from [1], pages 60-62.

  The sufficient decrease condition may be impossible to satisfy close to a
  minimum, in that case, we switch to an approximate sufficient decrease
  condition (approximate Wolfe) taken from [2].

  Supports complex variables, see [3].

  [1] J. Nocedal and S. Wright, 'Numerical Optimization', 2nd edition, 2006.
  [2] W. Hager, H. Zhang, Algorithm 851: CG_DESCENT, a Conjugate Gradient Method
    with Guaranteed Descent.
  [3] L. Sorber, M. van Barel, and L. de Lathauwer, 'Unconstrained Optimization
   of Real Functions in Complex Variables', SIAM J. Optim., Vol. 22, No. 3, pp. 879-898

  Attributes:
    fun: a function of the form ``fun(params, *args, **kwargs)``, where
      ``params`` are parameters of the model, ``*args`` and ``**kwargs`` are
      additional arguments.
    value_and_grad: if ``False``, ``fun`` should return the function value only.
      If ``True``, ``fun`` should return both the function value and the
      gradient.
    has_aux: if ``False``, ``fun`` should return the function value only. If
      ``True``, ``fun`` should return a pair ``(value, aux)`` where ``aux`` is a
      pytree of auxiliary values. (default: False)
    c1: constant used to check if a sufficient decrease has been found (Armijo)
      (default: 1e-4)
    c2: constant used to check if a small curvature has been found (strong
      Wolfe) (default: 0.9)
    c3: constant used to check if an approximate sufficient decrease
      (approximate Wolfe) has been found (default: 1e-6)
    rel_tol_cubic: point computed by cubic interpolation accepted if inside
      rel_tol_cubic*interval_size (default: 0.2)
    rel_tol_quad: point computed by quadratic interpolation accepted if inside
      rel_tol_quad*interval_size (default: 0.1)
    interval_threshold: if the size of the interval searched is below this threshold
      and a sufficient decrease for some stepsize s has been found, 
      then the linesearch takes s and moves on. (default=5*1e-5, which corresponds 
      to 15 linesearch iterations for init_stepsize = 1)
    increase_factor: factor to mutliply stepsize at initialization until finding
      interval satisfying curvature condition (default: 2.)
    max_stepsize: maximal possible stepsize, (default: 2**maxiter)
    tol: tolerance of the stopping criterion. (default: 0.0)
    maxiter: maximum number of line search iterations. (default: 30)
    verbose: whether to print information on every iteration or not.
    jit: whether to JIT-compile the optimization loop (default: "auto").
    unroll: whether to unroll the optimization loop (default: "auto").
  """

  fun: Callable
  value_and_grad: Union[bool, Callable] = False
  has_aux: bool = False

  c1: float = 1e-4
  c2: float = 0.9
  c3: float = 1e-6
  rel_tol_cubic: float = 0.2
  rel_tol_quad: float = 0.1
  interval_threshold: float = 1e-6
  increase_factor: float = 2.0

  tol: float = 0.0
  maxiter: int = 30
  max_stepsize: Optional[float] = None

  verbose: Union[bool, int] = False
  jit: base.AutoOrBoolean = "auto"
  unroll: base.AutoOrBoolean = "auto"

  def _value_and_slope_on_line(
      self, params, stepsize, descent_direction, args, kwargs
  ):
    step = tree_add_scalar_mul(params, stepsize, descent_direction)
    (value_step, aux_step), grad_step = self._value_and_grad_fun_with_aux(
        step, *args, **kwargs
    )
    slope_step = tree_vdot_real(tree_conj(grad_step), descent_direction)
    return value_step, slope_step, step, grad_step, aux_step

  def _decrease_error(
      self, stepsize, value_step, slope_step, value_init, slope_init
  ):
    # We consider either the usual sufficient decrease (Armijo condition), see
    # equation (3.7a) of [1]
    exact_decrease_error = (
        value_step - value_init - self.c1 * stepsize * slope_init
    )
    # or an approximate decrease condition, see equation (23) of [2]
    approx_decrease_error = slope_step - (2 * self.c1 - 1.0) * slope_init

    # The classical Armijo condition may fail to be satisfied if we are too
    # close to a minimum, causing the optimizer to fail as explained in [2]

    # We switch to approximate Wolfe conditions only if we are close enough to
    # the minimizer which is captured by the following criterion.
    delta_values = value_step - value_init - self.c3 * jnp.abs(value_init)
    approx_decrease_error = jnp.maximum(approx_decrease_error, delta_values)
    # We take then the *minimum* of both errors.
    return jnp.minimum(approx_decrease_error, exact_decrease_error)

  def _curvature_error(self, slope_step, slope_init):
    # See equation (3.7b) of [1].
    return jnp.abs(slope_step) - self.c2 * jnp.abs(slope_init)

  def _make_safe_step(self, stepsize, state, args, kwargs):
    safe_stepsize = state.safe_stepsize
    outside_domain = jnp.isinf(state.decrease_error)
    final_stepsize = jnp.where((safe_stepsize > 0.) | outside_domain, safe_stepsize, stepsize)
    if self.verbose:
      jax.lax.cond(
        safe_stepsize > 0.,
        lambda *_: jax.debug.print(FLAG_CURVATURE_COND_NOT_SATSIFIED),
        self.failure_diagnostic,
        stepsize,
        state
      )

    step = tree_add_scalar_mul(
        state.params, final_stepsize, state.descent_direction
    )
    (value_step, aux_step), grad_step = self._value_and_grad_fun_with_aux(
        step, *args, **kwargs
    )
    new_state = state._replace(
        params=step,
        value=value_step,
        grad=grad_step,
        aux=aux_step,
    )
    return final_stepsize, new_state

  def _keep_step(self, stepsize, state, _, __):
    return stepsize, state

  def _search_interval(self, init_stepsize, state, args, kwargs):
    """Line search procedure described in Algorithm 3.5 of [1]."""
    # init_stepsize only used for iter_num = 0

    iter_num = state.iter_num

    params_init = state.params
    grad_init = state.grad
    aux_init = state.aux

    value_init = state.value_init
    slope_init = state.slope_init
    descent_direction = state.descent_direction

    prev_stepsize = state.prev_stepsize
    prev_value_step = state.prev_value_step
    prev_slope_step = state.prev_slope_step

    safe_stepsize = state.safe_stepsize

    # Choose new point, larger than previous one or set to initial guess
    # for first iteration.
    larger_stepsize = self.increase_factor * prev_stepsize
    new_stepsize = jnp.where(iter_num == 0, init_stepsize, larger_stepsize)
    new_stepsize = jnp.minimum(new_stepsize, self.max_stepsize)
    max_stepsize_reached = new_stepsize >= self.max_stepsize

    new_value_step, new_slope_step, new_step, new_grad_step, new_aux_step = (
        self._value_and_slope_on_line(
            params_init, new_stepsize, descent_direction, args, kwargs
        )
    )
    is_value_nan = jnp.isnan(new_value_step) | jnp.isinf(new_value_step)
    if self.verbose:
      _cond_print(is_value_nan, FLAG_NAN_INF_VALUES)

    decrease_error = self._decrease_error(
        new_stepsize, new_value_step, new_slope_step, value_init, slope_init
    )
    decrease_error = jnp.maximum(decrease_error, 0.0)
    decrease_error = jnp.where(
        jnp.isnan(decrease_error), jnp.inf, decrease_error
    )

    curvature_error = self._curvature_error(new_slope_step, slope_init)
    curvature_error = jnp.maximum(curvature_error, 0.0)
    curvature_error = jnp.where(
        jnp.isnan(curvature_error), jnp.inf, curvature_error
    )
    new_error = jnp.maximum(decrease_error, curvature_error)

    # If the new point satisfies at least the decrease error we keep it
    # in case the curvature error cannot be satisfied.
    safe_decrease = decrease_error <= self.tol
    new_safe_stepsize = jnp.where(safe_decrease, new_stepsize, safe_stepsize)

    # If the new point not good, set high and low values according to
    # conditions described in Algorithm 3.5 of [1]
    set_high_to_new = (decrease_error > 0.0) | (
        (new_value_step >= prev_value_step) & (iter_num > 0)
    )
    set_low_to_new = (new_slope_step >= 0.0) & (~set_high_to_new)

    # By default we set high to new and correct if we should have set
    # low to new. If none should have set, the search for the interval
    # continues anyway.
    low_, value_low_, slope_low_, high_, value_high_, slope_high_ = (
        prev_stepsize,
        prev_value_step,
        prev_slope_step,
        new_stepsize,
        new_value_step,
        new_slope_step,
    )

    default = [low_, value_low_, slope_low_, high_, value_high_, slope_high_]
    candidate = [
        new_stepsize,
        new_value_step,
        new_slope_step,
        prev_stepsize,
        prev_value_step,
        prev_slope_step,
    ]
    [low, value_low, slope_low, high, value_high, slope_high] = _set_values(
        set_low_to_new, candidate, default
    )

    # If high or low have been set or the point is good, the interval has been
    # found. Otherwise we'll keep on augmenting the stepsize.
    interval_found = set_high_to_new | set_low_to_new | (new_error <= self.tol)

    # If new_error <= self.tol, the line search is done. If the maximal stepsize
    # is reached, either an interval has been found and we will zoom into this
    # interval or no interval has been found meaning that the maximal stepsize
    # satisfies the Armijo condition but a priori not the curvature condition.
    # In that case there is no hope to find the curvature condition as it would
    # a priori be found for a larger stepsize, so we simply take the maximal
    # stepsize and flag that we could not satisfy the curvature condition. If
    # the linesearch is done for either of the two reasons above, we set
    # directly the new parameters, gradient, value and aux to the ones found.
    done = (new_error <= self.tol) | (max_stepsize_reached & ~interval_found)
    default = [params_init, value_init, grad_init, aux_init]
    candidate = [
        new_step,
        new_value_step,
        new_grad_step,
        new_aux_step,
    ]
    next_params, next_value, next_grad, next_aux = _set_values(
        done, candidate, default
    )
    if self.verbose:
      _cond_print(
        (max_stepsize_reached & ~interval_found),
        FLAG_INTERVAL_NOT_FOUND + '\n' 
        + FLAG_CURVATURE_COND_NOT_SATSIFIED
      )
    max_iter_reached = (iter_num + 1 >= self.maxiter) & (~done)

    new_state = state._replace(
        iter_num=iter_num + 1,
        params=next_params,
        value=next_value,
        grad=next_grad,
        aux=next_aux,
        #
        decrease_error=decrease_error,
        curvature_error=curvature_error,
        error=new_error,
        done=done,
        failed=jnp.asarray(max_iter_reached),
        interval_found=interval_found,
        #
        prev_stepsize=new_stepsize,
        prev_value_step=new_value_step,
        prev_slope_step=new_slope_step,
        #
        low=low,
        value_low=value_low,
        slope_low=slope_low,
        high=high,
        value_high=value_high,
        slope_high=slope_high,
        cubic_ref=low,
        value_cubic_ref=value_low,
        #
        safe_stepsize=new_safe_stepsize,
    )
    return base.LineSearchStep(stepsize=new_stepsize, state=new_state)

  def _zoom_into_interval(self, stepsize, state, args, kwargs):
    """Zoom procedure described in Algorithm 3.6 of [1]."""

    # The stepsize is not used, only low, high, etc... are used to
    # find a good point
    dtype = stepsize.dtype
    del stepsize

    iter_num = state.iter_num

    params_init = state.params
    grad_init = state.grad
    aux_init = state.aux

    value_init = state.value_init
    slope_init = state.slope_init
    descent_direction = state.descent_direction

    low = state.low
    value_low = state.value_low
    slope_low = state.slope_low
    high = state.high
    value_high = state.value_high
    slope_high = state.slope_high
    cubic_ref = state.cubic_ref
    value_cubic_ref = state.value_cubic_ref

    safe_stepsize = state.safe_stepsize

    # Check if interval not too small otherwise fail
    delta = jnp.abs(high - low)
    left = jnp.minimum(high, low)
    right = jnp.maximum(high, low)
    cubic_chk = self.rel_tol_cubic * delta
    quad_chk = self.rel_tol_quad * delta

    # Rather large values of threshold compared to machine precision
    # such that we avoid wasting iterations to satisfy curvature condition
    # (a stepsize reducing values is taken if it exists when threshold is met)
    threshold = self.interval_threshold
    too_small_int = delta <= threshold
    if self.verbose:
      _cond_print(too_small_int, FLAG_INTERVAL_TOO_SMALL)

    # Find new point by interpolation
    middle_cubic = _cubicmin(
        low, value_low, slope_low, high, value_high, cubic_ref, value_cubic_ref
    )
    middle_cubic_valid = (middle_cubic > left + cubic_chk) & (
        middle_cubic < right - cubic_chk
    )
    use_cubic = middle_cubic_valid
    middle_quad = _quadmin(low, value_low, slope_low, high, value_high)
    middle_quad_valid = (middle_quad > left + quad_chk) & (
        middle_quad < right - quad_chk
    )
    use_quad = (~use_cubic) & middle_quad_valid
    middle_bisection = (low + high) / 2.0
    use_bisection = (~use_cubic) & (~use_quad)

    middle = jnp.where(use_cubic, middle_cubic, cubic_ref)
    middle = jnp.where(use_quad, middle_quad, middle)
    middle = jnp.where(use_bisection, middle_bisection, middle).astype(dtype)

    # Check if new point is good
    value_middle, slope_middle, step, grad_step, aux_step = (
        self._value_and_slope_on_line(
            params_init, middle, descent_direction, args, kwargs
        )
    )
    is_value_nan = jnp.isnan(value_middle) | jnp.isinf(value_middle)
    if self.verbose:
      _cond_print(is_value_nan, FLAG_NAN_INF_VALUES)

    decrease_error = self._decrease_error(
        middle, value_middle, slope_middle, value_init, slope_init
    )
    decrease_error = jnp.maximum(decrease_error, 0.0)
    decrease_error = jnp.where(
        jnp.isnan(decrease_error), jnp.inf, decrease_error
    )

    curvature_error = self._curvature_error(slope_middle, slope_init)
    curvature_error = jnp.maximum(curvature_error, 0.0)
    curvature_error = jnp.where(
        jnp.isnan(curvature_error), jnp.inf, curvature_error
    )

    new_error = jnp.maximum(decrease_error, curvature_error)

    # If the new point satisfies at least the decrease error we keep it in case
    # the curvature error cannot be satisfied. We take the largest possible one
    safe_decrease = decrease_error <= self.tol
    new_safe_stepsize = jnp.where(safe_decrease, middle, safe_stepsize)
    new_safe_stepsize = jnp.maximum(new_safe_stepsize, safe_stepsize)

    # If both armijo and curvature conditions are satisfied, we are done.
    done = new_error <= self.tol
    default = [params_init, value_init, grad_init, aux_init]
    candidate = [step, value_middle, grad_step, aux_step]
    next_params, next_value, next_grad, next_aux = _set_values(
        new_error <= self.tol, candidate, default
    )

    # Otherwise, we update high and low values
    set_high_to_middle = (decrease_error > 0.0) | (value_middle >= value_low)
    secant_interval = slope_middle * (high - low)
    set_high_to_low = (secant_interval >= 0.0) & (~set_high_to_middle)
    set_low_to_middle = ~set_high_to_middle

    # Set high to middle, or low, or keep as it is
    default = [high, value_high, slope_high]
    candidate = [middle, value_middle, slope_middle]
    [new_high_, new_value_high_, new_slope_high_] = _set_values(
        set_high_to_middle, candidate, default
    )
    default = [new_high_, new_value_high_, new_slope_high_]
    candidate = [low, value_low, slope_low]
    [new_high, new_value_high, new_slope_high] = _set_values(
        set_high_to_low, candidate, default
    )

    # Set low to middle or keep as it is
    default = [low, value_low, slope_low]
    candidate = [middle, value_middle, slope_middle]
    [new_low, new_value_low, new_slope_low] = _set_values(
        set_low_to_middle, candidate, default
    )

    # Update cubic reference point.
    # If high changed then it can be used as the new ref point.
    # Otherwise, low has been updated and not kept as high
    # so it can be used as the new ref point.
    [new_cubic_ref, new_value_cubic_ref] = _set_values(
        set_high_to_middle | set_high_to_low,
        [high, value_high],
        [low, value_low],
    )
    # We stop if the searched interval is reduced below machine precision 
    # and we already have found a positive stepsize ensuring sufficient
    # decrease. If no stepsize with sufficient decrease has been found, 
    # we keep going on (some extremely steep functions require very small
    # stepsizes, see zakharov test in lbfgs_test.py)
    max_iter_reached = (iter_num + 1 >= self.maxiter)
    presumably_failed = jnp.asarray(max_iter_reached) | \
      (too_small_int & (new_safe_stepsize > 0.))
    failed = presumably_failed & ~done
    new_state = state._replace(
        iter_num=iter_num + 1,
        params=next_params,
        value=next_value,
        grad=next_grad,
        aux=next_aux,
        #
        decrease_error=decrease_error,
        curvature_error=curvature_error,
        error=new_error,
        done=done,
        failed=failed,
        #
        low=new_low,
        value_low=new_value_low,
        slope_low=new_slope_low,
        high=new_high,
        value_high=new_value_high,
        slope_high=new_slope_high,
        cubic_ref=new_cubic_ref,
        value_cubic_ref=new_value_cubic_ref,
        #
        safe_stepsize=new_safe_stepsize,
    )
    return base.LineSearchStep(stepsize=middle, state=new_state)

  def init_state(
      self,
      init_stepsize: float,
      params: Any,
      value: Optional[float] = None,
      grad: Optional[Any] = None,
      descent_direction: Optional[Any] = None,
      fun_args: list = [],
      fun_kwargs: dict = {},
  ) -> base.LineSearchStep:
    """Initialize the line search state by computing all relevant quantities and store it in the initial state.

    Args:
      init_stepsize: initial step size value (used in update, not in
        init_state).
      params: current parameters.
      value: current function value (recomputed if None).
      grad: current gradient (recomputed if None).
      descent_direction: descent direction (negative gradient if None).
      fun_args: additional positional arguments to be passed to ``fun``.
      fun_kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      state
    """
    # FIXME: Signature issue in base.IterativeLineSearch: Keyword argument
    # before variable positional arguments.
    realdtype = get_real_dtype(tree_single_dtype(params))
    num_fun_eval = jnp.asarray(0, base.NUM_EVAL_DTYPE)
    num_grad_eval = jnp.asarray(0, base.NUM_EVAL_DTYPE)
    del init_stepsize
    aux = None
    if value is None or grad is None:
      (value, aux), grad = self._value_and_grad_fun_with_aux(
          params, *fun_args, **fun_kwargs
      )
      num_fun_eval = num_fun_eval + 1
      num_grad_eval = num_grad_eval + 1

    # TODO(vroulet): ideally, we shall also provide aux as arguments to avoid
    # recomputing the function. It's especially problematic if the function
    # provided has an artificial aux = None coming from its instanciation via
    # base._make_funs_with_aux. This requires changing the signature of
    # base.IterativeLineSearch.
    if aux is None and self.has_aux:
      _, aux = self._fun_with_aux(params, *fun_args, **fun_kwargs)
      num_fun_eval += 1
      num_grad_eval += 1

    if descent_direction is None:
      descent_direction = tree_scalar_mul(-1.0, tree_conj(grad))

    slope = tree_vdot_real(tree_conj(grad), descent_direction)

    return ZoomLineSearchState(
        iter_num=jnp.asarray(0),
        params=params,
        value=value,
        grad=grad,
        aux=aux,
        #
        value_init=value,
        slope_init=slope,
        descent_direction=descent_direction,
        #
        decrease_error=jnp.asarray(jnp.inf),
        curvature_error=jnp.asarray(jnp.inf),
        error=jnp.asarray(jnp.inf),
        done=jnp.asarray(False),
        failed=jnp.asarray(False),
        interval_found=jnp.asarray(False),
        #
        prev_stepsize=jnp.asarray(0.0).astype(realdtype),
        prev_value_step=value,
        prev_slope_step=slope,
        #
        low=jnp.asarray(0.0).astype(realdtype),
        value_low=value,
        slope_low=slope,
        high=jnp.asarray(0.0).astype(realdtype),
        value_high=value,
        slope_high=slope,
        cubic_ref=jnp.asarray(0.0).astype(realdtype),
        value_cubic_ref=value,
        #
        safe_stepsize=jnp.asarray(0.0).astype(realdtype),
        num_fun_eval=num_fun_eval,
        num_grad_eval=num_grad_eval,
    )

  def update(
      self,
      stepsize: float,
      state: NamedTuple,
      params: Any,
      value: Optional[float] = None,
      grad: Optional[Any] = None,
      descent_direction: Optional[Any] = None,
      fun_args: list = [],
      fun_kwargs: dict = {},
  ) -> base.LineSearchStep:
    """Combines Algorithms 3.5 and 3.6 of [1].

    Final state contains next_params, next_value, next_grad, next_aux if the
    linesearch succeeded.

    Args:
      stepsize: current estimate of the step size.
      state: named tuple containing the line search state.
      params: current parameters (not used, recorded during init in state).
      value: current function value (not used, recorded during init in state).
      grad: current gradient (not used, recorded during init in state).
      descent_direction: descent direction (not used, recorded during init in
        state).
      fun_args: additional positional arguments to be passed to ``fun``.
      fun_kwargs: additional keyword arguments to be passed to ``fun``.

    Returns:
      (stepsize, state)
    """
    # FIXME: Signature issue in base.IterativeLineSearch: Keyword argument
    # before variable positional arguments.
    # Params, value, grad, descent direction recorded in state at initialization
    realdtype = get_real_dtype(tree_single_dtype(params))
    init_stepsize = jnp.asarray(stepsize).astype(realdtype)
    del params
    del value
    del grad
    del descent_direction

    new_stepsize, new_state = cond(
        state.interval_found,
        self._zoom_into_interval,
        self._search_interval,
        init_stepsize,
        state,
        fun_args,
        fun_kwargs,
        jit=self.jit,
    )
    new_state = new_state._replace(
        num_fun_eval=new_state.num_fun_eval + 1,
        num_grad_eval=new_state.num_grad_eval + 1,
    )

    anticipated_num_func_grad_calls = jnp.array(
      new_state.failed
    ).astype(base.NUM_EVAL_DTYPE)
    new_stepsize, new_state = cond(
        new_state.failed,
        self._make_safe_step,
        self._keep_step,
        new_stepsize,
        new_state,
        fun_args,
        fun_kwargs,
        jit=self.jit,
    )
    new_state = new_state._replace(
        num_fun_eval=new_state.num_fun_eval + anticipated_num_func_grad_calls,
        num_grad_eval=new_state.num_grad_eval + anticipated_num_func_grad_calls,
    )

    if self.verbose:
      self._log_info(new_state, new_stepsize)

    return base.LineSearchStep(stepsize=new_stepsize, state=new_state)

  def _cond_fun(self, inputs):
    # Stop the linesearch according to done or failed rather than the error as one may
    # reach the maximal stepsize and no decrease of the curvature error may be
    # possible or the searched interval has been reduced too much.
    _, state = inputs[0]
    return ~(state.done | state.failed)

  def _log_info(self, state, stepsize):
    self.log_info(
        state,
        error_name="Minimum Decrease & Curvature Errors",
        additional_info={
          "Stepsize": stepsize,
          "Decrease Error": state.decrease_error,
          "Curvature Error": state.curvature_error
        }
    )

  def failure_diagnostic(self, stepsize, state):
    jax.debug.print(FLAG_NO_STEPSIZE_FOUND)
    self._log_info(state, stepsize)

    slope_init = state.slope_init
    is_descent_dir = slope_init < 0.
    _cond_print(
      ~is_descent_dir,
      FLAG_NOT_A_DESCENT_DIRECTION + \
      "The slope (={slope_init}) at stepsize=0 should be negative",
      slope_init=slope_init
    )
    _cond_print(
      is_descent_dir,
      WARNING_PREAMBLE + \
      "Consider augmenting the maximal number of linesearch iterations."
    )
    eps = jnp.finfo(stepsize).eps
    below_eps = stepsize < eps
    _cond_print(
      below_eps & is_descent_dir,
      WARNING_PREAMBLE + \
      "Computed stepsize (={stepsize}) " + \
      "is below machine precision (={eps}), " +\
      "consider passing to higher precision like x64, using " +\
      "jax.config.update('jax_enable_x64', True).",
      stepsize=stepsize, 
      eps=eps
    )
    abs_slope_init = jnp.abs(slope_init)
    high_slope = abs_slope_init > 1e16
    _cond_print(
      high_slope & is_descent_dir,
      WARNING_PREAMBLE + \
      "Very large absolute slope at stepsize=0. (|slope|={abs_slope_init}). " + \
      "The objective is badly conditioned. " + \
      "Consider reparameterizing objective (e.g., normalizing parameters) " + \
      "or finding a better guess for the initial parameters for the solver.",
      abs_slope_init=abs_slope_init
    )
    outside_domain = jnp.isinf(state.decrease_error)
    _cond_print(
      outside_domain,
      WARNING_PREAMBLE + \
      "Cannot even make a step without getting Inf or Nan. " + \
      "The linesearch won't make a step and the optimizer is stuck."
    )
    _cond_print(
      ~outside_domain,
      WARNING_PREAMBLE + \
      "Making an unsafe step, not decreasing enough the objective. " + \
      "Convergence of the solver is compromised as it does not reduce values."
    )

  def __post_init__(self):
    self._fun_with_aux, _, self._value_and_grad_fun_with_aux = (
        _make_funs_with_aux(
            self.fun, value_and_grad=self.value_and_grad, has_aux=self.has_aux
        )
    )
    if not self.max_stepsize:
      self.max_stepsize = float(2**self.maxiter)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Optional

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import MirrorDescent
from nemos.third_party.jaxopt.jaxopt import objective
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt import tree_util as tu

from sklearn import datasets
from sklearn import preprocessing


def kl_projection(x: Any, hyperparams_proj: Optional[Any] = None) -> Any:
  del hyperparams_proj
  return tu.tree_map(lambda t: jax.nn.softmax(t, -1), x)


# Generating function of the Bregman divergence.
kl_generating_fun = lambda x: -jnp.sum(jax.scipy.special.entr(x) + x)
# Row-wise mirror map.
kl_mapping_fun = jax.vmap(jax.grad(kl_generating_fun))


def projection_grad_kl_stable(x: Any,
                              x_fun_grad: Any,
                              stepsize: float,
                              hyperparams_proj: Optional[Any] = None) -> Any:
  del hyperparams_proj
  # TODO(fllinares): generalize to arbitrary axes. How to pass axes?
  def _fn(x_i, g_i):
    g_i = jnp.where(x_i != 0, -stepsize * g_i, -jnp.inf)
    g_i = g_i - jnp.max(g_i, axis=-1, keepdims=True)
    y_i = x_i * jnp.exp(g_i)
    return y_i / jnp.sum(y_i, axis=-1, keepdims=True)
  return tu.tree_map(_fn, x, x_fun_grad)


def make_stepsize_schedule(max_stepsize, n_steps, power=1.0) -> Callable:
  def stepsize_schedule(t: int) -> float:
    true_fn = lambda t: 1.0
    false_fn = lambda t: (n_steps / t) ** power
    decay_factor = jax.lax.cond(t <= n_steps, true_fn, false_fn, t)
    return decay_factor * max_stepsize
  return stepsize_schedule


class MirrorDescentTest(test_util.JaxoptTestCase):

  @parameterized.named_parameters(
      # FIXME: Re-enable when JAX numerical stability issue is solved.
      # ('kl', None),
      ('kl_stable', projection_grad_kl_stable),
  )
  def test_multiclass_svm_dual(self, projection_grad):
    X, y = datasets.make_classification(
        n_samples=20,
        n_features=5,
        n_informative=3,
        n_classes=3,
        random_state=0)
    Y = preprocessing.LabelBinarizer().fit_transform(y)
    Y = jnp.asarray(Y)
    n_samples, n_classes = Y.shape
    fun = objective.multiclass_linear_svm_dual
    lam = 10.0
    data = (X, Y)
    stepsize_schedule = make_stepsize_schedule(
        max_stepsize=5.0, n_steps=50, power=1.0)
    maxiter = 500
    tol = 1e-2
    atol = 1e-2

    beta_init = jnp.ones((n_samples, n_classes)) / n_classes
    if projection_grad is None:
      projection_grad = MirrorDescent.make_projection_grad(
          kl_projection, kl_mapping_fun)
    md = MirrorDescent(
        fun=fun,
        projection_grad=projection_grad,
        stepsize=stepsize_schedule,
        maxiter=maxiter,
        tol=tol,
        implicit_diff=True)
    beta_fit, info = md.run(beta_init, None, lam, data)

    # Check optimality conditions.
    self.assertLessEqual(info.error, tol)

    # Compare against sklearn.
    W_skl = test_util.multiclass_linear_svm_skl(X, y, lam)
    W_fit = jnp.dot(X.T, (Y - beta_fit)) / lam
    self.assertArraysAllClose(W_fit, W_skl, atol=atol)

  @parameterized.named_parameters(
      # FIXME: Re-enable when JAX numerical stability issue is solved.
      # ('kl', None),
      ('kl_stable', projection_grad_kl_stable),
  )
  def test_multiclass_svm_dual_implicit_diff(self, projection_grad):
    X, y = datasets.make_classification(
        n_samples=20,
        n_features=5,
        n_informative=3,
        n_classes=3,
        random_state=0)

    X = preprocessing.Normalizer().fit_transform(X)
    Y = preprocessing.LabelBinarizer().fit_transform(y)
    Y = jnp.asarray(Y)

    n_samples, n_classes = Y.shape
    fun = objective.multiclass_linear_svm_dual
    lam = 10.0
    data = (X, Y)
    stepsize_schedule = make_stepsize_schedule(
        max_stepsize=5.0, n_steps=50, power=1.0)
    maxiter = 500
    tol = 1e-3

    beta_init = jnp.ones((n_samples, n_classes)) / n_classes

    if projection_grad is None:
      projection_grad = MirrorDescent.make_projection_grad(
          kl_projection, kl_mapping_fun)

    md = MirrorDescent(
        fun=fun,
        projection_grad=projection_grad,
        stepsize=stepsize_schedule,
        maxiter=maxiter,
        tol=tol,
        implicit_diff=True)

    def mirror_descent_fun_primal(lam):
      beta_fit = md.run(beta_init, None, lam, data).params
      return jnp.dot(X.T, (Y - beta_fit)) / lam

    jac_primal = jax.jacrev(mirror_descent_fun_primal)(lam)
    jac_num = test_util.multiclass_linear_svm_skl_jac(X, y, lam, eps=1e-3)
    self.assertArraysAllClose(jac_num, jac_primal, atol=5e-3)

    def mirror_descent_fun_primal(lam):
      beta_fit = md.run(beta_init, l2reg=lam, data=data).params
      return jnp.dot(X.T, (Y - beta_fit)) / lam

    jac_primal = jax.jacrev(mirror_descent_fun_primal)(lam)
    jac_num = test_util.multiclass_linear_svm_skl_jac(X, y, lam, eps=1e-3)
    self.assertArraysAllClose(jac_num, jac_primal, atol=5e-3)


if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



import jax.numpy as jnp
import numpy as np
import pytest

from nemos import tree_utils


@pytest.mark.parametrize(
    "array, expected",
    [
        (jnp.array([[1, 2], [3, 4]]), jnp.array([True, True])),
        (jnp.array([[1, jnp.inf], [3, 4]]), jnp.array([False, True])),
        (jnp.array([[1, 2], [jnp.inf, jnp.inf]]), jnp.array([True, False])),
    ],
)
def test_get_not_inf(array, expected):
    """Test _get_not_inf function for correctly identifying non-infinite values."""
    assert jnp.array_equal(tree_utils._get_not_inf(array), expected)


@pytest.mark.parametrize(
    "array, expected",
    [
        (jnp.array([[1, 2], [3, 4]]), jnp.array([True, True])),
        (jnp.array([[1, jnp.nan], [3, 4]]), jnp.array([False, True])),
        (jnp.array([[1, 2], [jnp.nan, jnp.nan]]), jnp.array([True, False])),
    ],
)
def test_get_not_nan(array, expected):
    """Test _get_not_nan function for correctly identifying non-NaN values."""
    assert jnp.array_equal(tree_utils._get_not_nan(array), expected)


@pytest.mark.parametrize(
    "tree, expected_shape",
    [
        (jnp.array([[1], [2], [3], [4]]), 4),
        ({"x": {"y": jnp.array([1, 2, jnp.nan])}}, 3),
    ],
)
def test_check_valid_length(tree, expected_shape):
    """Test that validation of trees returns an array of the right first shape."""
    valid = tree_utils._get_valid_tree(tree)
    assert valid.shape[0] == expected_shape


@pytest.mark.parametrize(
    "tree",
    [(jnp.array([[1], [2], [3], [4]])), ({"x": {"y": jnp.array([1, 2, jnp.nan])}})],
)
def test_check_flat_array(tree):
    """Test that validation of trees returns an array of the right dimensionality."""
    valid = tree_utils._get_valid_tree(tree)
    assert valid.ndim == 1


@pytest.mark.parametrize(
    "tree, expected",
    [
        (
            {"a": jnp.array([[1, 2], [3, 4]]), "b": jnp.array([[5, 6], [7, 8]])},
            jnp.array([True, True]),
        ),
        (
            {"a": jnp.array([[1, 2], [jnp.nan, 4]]), "b": jnp.array([[5, 6], [7, 8]])},
            jnp.array([True, False]),
        ),
        (
            {
                "a": jnp.array([[1, jnp.nan], [3, 4]]),
                "b": jnp.array([[5, 6], [jnp.inf, 8]]),
            },
            jnp.array([False, False]),
        ),
    ],
)
def test_get_valid_tree(tree, expected):
    """Test _get_valid_tree function for filtering valid tree entries."""
    assert jnp.array_equal(tree_utils._get_valid_tree(tree), expected)


@pytest.mark.parametrize(
    "trees, expected",
    [
        (
            ({"a": jnp.array([[1, 2], [3, 4]])}, {"b": jnp.array([[5, 6], [7, 8]])}),
            jnp.array([True, True]),
        ),
        (
            (
                {"a": jnp.array([[1, 2], [3, 4]])},
                {"b": jnp.array([[5, 6], [jnp.nan, 8]])},
            ),
            jnp.array([True, False]),
        ),
        (
            (
                {"a": jnp.array([[1, jnp.nan], [3, 4]])},
                {"b": jnp.array([[4, 6], [7, jnp.inf]])},
            ),
            jnp.array([False, False]),
        ),
    ],
)
def test_get_valid_multitree(trees, expected):
    """Test get_valid_multitree function for filtering valid entries across multiple trees."""
    assert jnp.array_equal(tree_utils.get_valid_multitree(*trees), expected)


@pytest.mark.parametrize(
    "idx",
    [
        slice(2, 5),  # Slice indexing
        np.array([1, 3, 5]),  # Integer list indexing
        np.array(
            [True, False, True, False, True, False, True, False, True, False]
        ),  # Boolean array indexing
        (slice(1, 3), slice(0, 2)),  # Mixed indexing (simple example with slices)
    ],
)
def test_tree_slice(idx):
    mydict = {
        "array1": np.random.rand(10, 3),
        "array2": np.random.rand(10, 2),
        "array3": np.random.rand(10, 4),
        "array4": jnp.arange(30).reshape(10, 3),
    }
    result = tree_utils.tree_slice(mydict, idx)
    for key in mydict:
        expected = mydict[key][idx]
        assert jnp.all(result[key] == expected)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp

from nemos.third_party.jaxopt.jaxopt import loop
from nemos.third_party.jaxopt.jaxopt._src import test_util


class LoopTest(test_util.JaxoptTestCase):

  @parameterized.product(unroll=[True, False], jit=[True, False])
  def test_while_loop(self, unroll, jit):
    def my_pow(x, y):
      def body_fun(val):
        return val * x
      def cond_fun(val):
        return True
      return loop.while_loop(cond_fun=cond_fun, body_fun=body_fun, init_val=1.0,
                             maxiter=y, unroll=unroll, jit=jit)

    if not unroll and not jit:
      self.assertRaises(ValueError, my_pow, 3, 4)
      return

    self.assertEqual(my_pow(3, 4), pow(3, 4))

    if unroll:
      # unroll=False uses lax.while_loop, whichs is not differentiable.
      self.assertEqual(jax.grad(my_pow)(3.0, 4),
                       jax.grad(jnp.power)(3.0, 4))

  @parameterized.product(unroll=[True, False], jit=[True, False])
  def test_while_loop_stopped(self, unroll, jit):
    def my_pow(x, y, max_val):
      def body_fun(val):
        return val * x
      def cond_fun(val):
        return val < max_val
      return loop.while_loop(cond_fun=cond_fun, body_fun=body_fun, init_val=1.0,
                             maxiter=y, unroll=unroll, jit=jit)

    if not unroll and not jit:
      self.assertRaises(ValueError, my_pow, 3, 4, max_val=81)
      return

    # We asked for pow(3, 6) but due to max_val, we get pow(3, 4).
    self.assertEqual(my_pow(3, 6, max_val=81), pow(3, 4))

    if unroll:
      self.assertEqual(jax.grad(my_pow)(3.0, 6, max_val=81),
                       jax.grad(jnp.power)(3.0, 4))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Line searches utilities."""

from jax import numpy as jnp
from nemos.third_party.jaxopt.jaxopt._src import base
from nemos.third_party.jaxopt.jaxopt._src.backtracking_linesearch import BacktrackingLineSearch
from nemos.third_party.jaxopt.jaxopt._src.hager_zhang_linesearch import HagerZhangLineSearch
from nemos.third_party.jaxopt.jaxopt._src.zoom_linesearch import ZoomLineSearch


def _setup_linesearch(
    linesearch,
    fun,
    value_and_grad,
    has_aux,
    maxlsiter,
    max_stepsize,
    jit,
    unroll,
    verbose,
):
  """Instantiate linesearch."""

  available_linesearches = ["backtracking", "zoom", "hager-zhang"]
  if linesearch == "backtracking":
    linesearch_solver = BacktrackingLineSearch(
        fun=fun,
        value_and_grad=value_and_grad,
        has_aux=has_aux,
        maxiter=maxlsiter,
        max_stepsize=max_stepsize,
        jit=jit,
        unroll=unroll,
        verbose=verbose,
    )
  elif linesearch == "zoom":
    linesearch_solver = ZoomLineSearch(
        fun=fun,
        value_and_grad=value_and_grad,
        has_aux=has_aux,
        maxiter=maxlsiter,
        max_stepsize=max_stepsize,
        jit=jit,
        unroll=unroll,
        verbose=verbose,
    )
  elif linesearch == "hager-zhang":
    # NOTE(vroulet): max_stepsize has no effect in HZ
    linesearch_solver = HagerZhangLineSearch(
        fun=fun,
        value_and_grad=value_and_grad,
        has_aux=has_aux,
        maxiter=maxlsiter,
        jit=jit,
        unroll=unroll,
        verbose=verbose,
    )
  elif isinstance(linesearch, base.IterativeLineSearch):
    linesearch_solver = linesearch
  else:
    raise ValueError(
        f"Linesearch {linesearch} not available/tested. "
        f"Available linesearches: {available_linesearches}"
    )
  return linesearch_solver


def _init_stepsize(
    strategy, max_stepsize, min_stepsize, increase_factor, stepsize
):
  """Set stepsize at the start of the linesearch from previous guess."""
  available_strategies = ["max", "current", "increase"]
  if strategy == "max":
    init_stepsize = max_stepsize
  elif strategy == "current":
    init_stepsize = stepsize
  elif strategy == "increase":
    init_stepsize = jnp.where(
        stepsize <= min_stepsize,
        # If stepsize became too small, we restart it.
        max_stepsize,
        # Else, we increase a bit the previous one.
        stepsize * increase_factor,
    )
  else:
    raise ValueError(
        f"Strategy {strategy} not available/tested. "
        f"Available linesearches: {available_strategies}"
    )
  return init_stepsize



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp
from jax import scipy as jsp
from jax.test_util import check_grads

from nemos.third_party.jaxopt import jaxopt
from nemos.third_party.jaxopt.jaxopt.tree_util import tree_l2_norm
from nemos.third_party.jaxopt.jaxopt import AndersonAcceleration
from nemos.third_party.jaxopt.jaxopt._src import test_util

import numpy as onp


class AndersonAccelerationTest(test_util.JaxoptTestCase):

  def _make_random_affine_contractive_mapping(self, n):
    """Return a pair (M,b) where M is a contractive mapping"""
    key, subkey = jax.random.split(jax.random.PRNGKey(0))
    M = jax.random.uniform(subkey, shape=(n,n), minval=-1., maxval=1.)
    M = M + M.T  # random symmetric matrix
    eigv = jnp.linalg.eigvalsh(M)
    spectral_radius = jnp.abs(eigv[-1])
    eps = 1e-4
    M = M / (2*spectral_radius + eps)  # contractive mapping: lbda_max < 0.5
    key, subkey = jax.random.split(key)
    b = jax.random.uniform(subkey, shape=(n,))
    return M, b

  def test_geometric_decay(self):
    """Test convergence for geometric progression with common ratio r < 1."""
    def f(x):
      return jnp.array([0.9, 0.99, 0.999]) * x
    x0 = jnp.array([1., 0.4, -0.2])
    tol = 1e-5
    sol, state = AndersonAcceleration(f, history_size=5, maxiter=10*1000, ridge=1e-6, tol=tol).run(x0)
    self.assertLess(state.error, tol)
    sol_norm = tree_l2_norm(sol)
    big_tol = 1e-2  # we are forced to take high tol since convergence is very slow for 0.999
    self.assertLess(sol_norm, big_tol)

  def test_support_for_scalars(self):
    def f(x):
      return 0.5 * x + 0.5
    x0 = jnp.array(0.)
    tol = 1e-4
    sol, state = AndersonAcceleration(f, history_size=5, maxiter=10*1000, ridge=1e-6, tol=tol).run(x0)
    self.assertLess(state.error, tol)

  @parameterized.product(jit=[False,True])
  def test_sin_fixed_point(self, jit):
    """Test convergence for simple polynomials and sin.

    Also test the support of pytree in input/output.
    """
    if jit:
      def f(x):  # Another fixed point exists for x[0] : ~1.11
        return jnp.sin(x[0]) * (x[0] ** 2), (x[1][0] ** 3, -x[1][1])
    else:
      def f(x):  # Another fixed point exists for x[0] : ~1.11
        return onp.sin(x[0]) * (x[0] ** 2), (x[1][0] ** 3, -x[1][1])
    x0 = jnp.array([0.6, 0., -0.1]), (jnp.array([[0.7], [0.5]]), jnp.array([[-1., 0., 2., -0.4]])/3)
    tol = 1e-6
    sol, state = AndersonAcceleration(f, history_size=3, maxiter=100*1000, ridge=1e-6, tol=tol, jit=jit).run(x0)
    self.assertLess(state.error, tol)
    sol_norm = tree_l2_norm(sol)
    self.assertLess(sol_norm, tol)
    f_fixed_point = (jnp.zeros(3), (jnp.zeros((2,1)), jnp.zeros((1,4))))
    self.assertAllClose(sol, f_fixed_point)

  def test_cos_fixed_point(self):
    """Test convergence for cos fixed point (non-zero fixed point).

    Also test support for additional parameters.
    """
    def f(x, theta):
      return jnp.cos(x) + theta
    x0 = jnp.array([0.3, jnp.pi / 4])
    theta = jnp.array([0., jnp.pi / 2])
    f_fixed_point = jnp.array([0.73908614842288, jnp.pi / 2])
    tol = 1e-6
    sol, state = AndersonAcceleration(f, history_size=4, maxiter=100, ridge=1e-6, tol=tol).run(x0, theta)
    self.assertLess(state.error, tol)
    self.assertArraysAllClose(sol, f_fixed_point)

  def test_rotation_map(self):
    """On this example Banach theorem does not hold. However Anderson works."""
    def rot(x, theta, c):
      n = theta.shape[0]
      cosO, sinO = jnp.cos(theta), jnp.sin(theta)
      first_row = jnp.stack([cosO, -sinO], axis=-1).reshape([-1])
      second_row = jnp.stack([sinO, cosO], axis=-1).reshape([-1])
      blocks = jnp.stack([first_row, second_row], axis=0)
      blocks = jnp.split(blocks, n, axis=1)
      M = jnp.array(jsp.linalg.block_diag(*blocks))
      return M @ (x - c) + c
    theta = jnp.pi / jnp.array([2, 3, 4, 5, 6, 1.27, 1.73])  # rotation angles
    c = jnp.arange(1,2*theta.shape[0]+1)  # rotation center
    key = jax.random.PRNGKey(0)
    x0 = jax.random.uniform(key, shape=(len(theta)*2,))
    tol = 1e-4
    aa = AndersonAcceleration(rot, history_size=10, maxiter=400, ridge=1e-7, tol=tol)
    sol, state = aa.run(x0, theta, c)
    self.assertLess(state.error, tol)
    self.assertArraysAllClose(sol, c.astype(dtype=jnp.float32), atol=1e-3)

  @parameterized.product(implicit_diff=[False, True])
  def test_simple_grads(self, implicit_diff):
    """Test correctness of gradients on a simple function."""
    def f(x, theta):
      return jnp.cos(x) + theta
    x0 = jnp.array([0., jnp.pi / 2])
    theta = jnp.array([0., jnp.pi / 2])
    tol = 1e-6
    ridge = 1e-5
    aa = AndersonAcceleration(f, history_size=5, maxiter=100, ridge=ridge, tol=tol, implicit_diff=implicit_diff)
    def solve_run(args, kwargs):
      return aa.run(x0, *args, **kwargs)[0]
    check_grads(solve_run, args=([theta], {}), order=1, modes=['rev'], eps=1e-4)

  def test_grads_flat_landscape(self):
    """Test correctness of gradients on a problem challenging for finite difference"""
    def f(x, theta):
      return theta * x
    x0 = jnp.array([1.,-0.5])
    theta = jnp.array([0.05, 0.2])
    aa = AndersonAcceleration(f, history_size=5, maxiter=10*1000, ridge=1e-6, tol=1e-6)
    def solve_run(args, kwargs):
      return aa.run(x0, *args, **kwargs)[0]
    check_grads(solve_run, args=([theta], {}), order=1, modes=['rev'], eps=1e-4)

  def test_affine_contractive_mapping(self):
    """Test correctness on big affine contractive mapping."""
    n = 200
    M, b = self._make_random_affine_contractive_mapping(n)
    def f(x, M, b):
      return M @ x + b
    tol = 1e-6
    aa = AndersonAcceleration(f, history_size=5, maxiter=100, ridge=1e-6, tol=tol)
    x0 = jnp.zeros_like(b)
    sol, state = aa.run(x0, M, b)
    self.assertLess(state.error, tol)
    def solve_run(args, kwargs):
      return aa.run(x0, *args, **kwargs)[0]
    check_grads(solve_run, args=([M, b], {}), order=1, modes=['rev'], eps=1e-4)

  def test_has_aux(self):
    """Non regression test for issue #443."""
    def fun_with_aux(x):
      y = 0.5 * x + 0.5
      aux = x*10
      return y, aux
    x0 = jnp.array([1.,-0.5])
    aa = AndersonAcceleration(fun_with_aux, has_aux=True,
                              history_size=5, maxiter=10*1000,
                              ridge=1e-6, tol=1e-4)
    sol, state = aa.run(x0)
    self.assertLess(state.error, 1e-4)

  @parameterized.product(mixing_frequency=[1, 5, 10])
  def test_mixing_frequency(self, mixing_frequency):
    """Test mixing_frequency on affine contractive mapping."""
    n = 10
    M, b = self._make_random_affine_contractive_mapping(n)
    def f(x, M, b):
      return M @ x + b
    tol = 1e-6
    aa = AndersonAcceleration(f, history_size=5, mixing_frequency=mixing_frequency,
                              maxiter=100, ridge=1e-6, tol=tol)
    x0 = jnp.zeros_like(b)
    sol, state = aa.run(x0, M, b)
    self.assertLess(state.error, tol)

  @parameterized.product(ridge=[1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1.])
  def test_residuals_minimizer(self, ridge):
    m, n = 10, 20
    key = jax.random.PRNGKey(0)
    key, subkey = jax.random.split(key)
    G = jax.random.normal(subkey, shape=(n,m))
    GTG = G.T @ G
    places = 4
    sol = jaxopt._src.anderson.minimize_residuals(GTG, ridge)
    mu, alpha = sol[0], sol[1:]
    # constyraint satisfaction
    self.assertAlmostEqual(jnp.sum(alpha), 1., places=places)
    # optimality conditions
    def opt_alpha(x):
      res = G @ x
      return jnp.sum(res ** 2)
    opt_constraint = lambda x: (jnp.sum(x) - 1)**2
    opt_alpha_grad = jax.grad(opt_alpha)(alpha)
    opt_constraint_grad = jax.grad(opt_constraint)(alpha)
    dot = jnp.dot(opt_alpha_grad, opt_constraint_grad)
    self.assertAlmostEqual(dot, 0., places=places)

if __name__ == '__main__':
  # Uncomment the line below in order to run in float64.
  # jax.config.update("jax_enable_x64", True)
  absltest.main()



import inspect
import re
from pydoc import render_doc
from typing import Type

from ._solver_registry import solver_registry


def get_solver_documentation(solver: str | Type, show_help: bool = False) -> str:
    """
    Get the documentation of a specified solver, including accepted arguments and the docstring of its `__init__`.

    Parameters
    ----------
    solver:
        `solver` can be a string or a type (e.g. `nemos.solvers.JaxoptGradientDescent`).
        If `solver` is a string, the corresponding solver will be read from the solver registry.
    show_help:
        Instead of the docstring, show the full output that would be produced by `help(solver)`
        where `solver` is a type.

    Example
    -------
    >>> import nemos as nmo
    >>> print(nmo.solvers.get_solver_documentation("SVRG"))
    Showing docstring of nemos.solvers._svrg.WrappedSVRG.
    For potentially more info, use `show_help=True`.
    <BLANKLINE>
    Adapter for NeMoS's implementation of SVRG following the AbstractSolver interface.
    <BLANKLINE>
    Accepted arguments:
    -------------------
    - batch_size
    - fun
    - key
    - maxiter
    - stepsize
    - tol
    <BLANKLINE>
    SVRG's documentation:
    ...
    """
    if isinstance(solver, str):
        solver = solver_registry[solver]

    if show_help:
        return render_doc(solver, title="Help on %s")

    solver_class_path = f"{solver.__module__}.{solver.__name__}"
    intro = inspect.cleandoc(
        f"""
        Showing docstring of {solver_class_path}.
        For potentially more info, use `show_help=True`.
        """
    )
    solver_doc = inspect.getdoc(solver) or f"No documentation found for {solver}."
    doc = intro + "\n\n" + solver_doc

    # if it doesn't already have it, then expand the docs
    # with the __init__'s docstring'
    pattern = r"More info from (.+)\.__init__"
    if re.search(pattern, doc) is None:
        solver_init_doc_header = inspect.cleandoc(
            f"More info from {solver.__name__}.__init__"
        )
        solver_init_doc_header += "\n" + "-" * len(solver_init_doc_header)
        solver_init_doc = inspect.cleandoc(
            inspect.getdoc(solver.__init__) or "No __init__ documentation found."
        )
        solver_init_doc = solver_init_doc_header + "\n" + solver_init_doc

        doc += "\n\n" + solver_init_doc

    return doc



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
from nemos.third_party.jaxopt.jaxopt._src import test_util
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _init_stepsize
from nemos.third_party.jaxopt.jaxopt._src.linesearch_util import _setup_linesearch


class LinesearchTest(test_util.JaxoptTestCase):
  @parameterized.product(
      linesearch=["zoom", "backtracking", "hager-zhang"],
      use_gradient=[False, True],
  )
  def test_linesearch_complex_variables(self, linesearch, use_gradient):
    """Test that optimization over complex variable z = x + jy matches equivalent real case"""

    W = jnp.array([[1, -2], [3, 4], [-4 + 2j, 5 - 3j], [-2 - 2j, 6]])

    def C2R(z):
      return jnp.stack((z.real, z.imag)) if z is not None else None

    def R2C(x):
      return x[..., 0, :] + 1j * x[..., 1, :]

    def f(z):
      return W @ z

    def loss_complex(z):
      return jnp.sum(jnp.abs(f(z)) ** 1.5)

    def loss_real(zR):
      return loss_complex(R2C(zR))

    z0 = jnp.array([1 - 1j, 0 + 1j])

    common_args = dict(
        value_and_grad=False,
        has_aux=False,
        maxlsiter=3,
        max_stepsize=1,
        jit=True,
        unroll=False,
        verbose=False,
    )

    ls_R = _setup_linesearch(
        linesearch=linesearch,
        fun=loss_real,
        **common_args,
    )

    ls_C = _setup_linesearch(
        linesearch=linesearch,
        fun=loss_complex,
        **common_args,
    )

    ls_state = _init_stepsize(
        strategy="increase",
        max_stepsize=1e-1,
        min_stepsize=1e-3,
        increase_factor=2.0,
        stepsize=1e-2,
    )

    descent_direction = (
        -jnp.conj(jax.grad(loss_complex)(z0)) if use_gradient else None
    )

    stepsize_R, _ = ls_R.run(
        ls_state, params=C2R(z0), descent_direction=C2R(descent_direction)
    )
    stepsize_C, _ = ls_C.run(
        ls_state, params=z0, descent_direction=descent_direction
    )

    self.assertArraysAllClose(stepsize_R, stepsize_C)


if __name__ == "__main__":
  absltest.main()



"""Utility functions for coupling filter definition."""

from __future__ import annotations

from typing import Callable, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np
import scipy.stats as sts
from numpy.typing import NDArray

from . import convolve, validation
from .pytrees import FeaturePytree


def difference_of_gammas(
    ws: int,
    upper_percentile: float = 0.99,
    inhib_a: float = 1.0,
    excit_a: float = 2.0,
    inhib_b: float = 1.0,
    excit_b: float = 2.0,
) -> NDArray:
    r"""
    Generate coupling filter as a Gamma pdf difference.

    Parameters
    ----------
    ws:
        The window size of the filter.
    upper_percentile:
        Upper bound of the gamma range as a percentile. The gamma function
        will be evaluated over the range [0, ppf(upper_percentile)].
    inhib_a:
        The ``a`` constant for the gamma pdf of the inhibitory part of the filter.
    excit_a:
        The ``a`` constant for the gamma pdf of the excitatory part of the filter.
    inhib_b:
        The ``b`` constant for the gamma pdf of the inhibitory part of the filter.
    excit_b:
        The ``a`` constant for the gamma pdf of the excitatory part of the filter.

    Notes
    -----
    The probability density function of a gamma distribution is parametrized as
    follows [1]_ :,

    .. math::

        p(x;\; a, b) = \frac{b^a x^{a-1} e^{-x}}{\Gamma(a)},

    where :math:`\Gamma(a)` refers to the gamma function, see [1]_.

    Returns
    -------
    filter:
        The coupling filter.

    Raises
    ------
    ValueError:
        If any of the Gamma parameters is lesser or equal to 0.
    ValueError:
        If the upper_percentile is not in [0, 1).

    References
    ----------
    .. [1] SciPy Docs - :obj:`scipy.stats.gamma`

    Examples
    --------
    .. plot::
        :include-source: True
        :caption: Difference of Gammas.

        >>> import matplotlib.pyplot as plt
        >>> from nemos.simulation import difference_of_gammas
        >>> coupling_duration = 100
        >>> inhib_a, inhib_b = 1.0, 1.0
        >>> excit_a, excit_b = 2.0, 2.0
        >>> coupling_filter = difference_of_gammas(
        ...     ws=coupling_duration,
        ...     inhib_a=inhib_a,
        ...     inhib_b=inhib_b,
        ...     excit_a=excit_a,
        ...     excit_b=excit_b
        ... )
        >>> _ = plt.plot(coupling_filter)
        >>> _ = plt.title("Coupling filter from difference of gammas")
        >>> _ = plt.show()

    """
    # check that the gamma parameters are positive (scipy returns
    # nans otherwise but no exception is raised)
    variables = {
        "excit_a": excit_a,
        "inhib_a": inhib_a,
        "excit_b": excit_b,
        "inhib_b": inhib_b,
    }
    for name, value in variables.items():
        if value <= 0:
            raise ValueError(f"Gamma parameter {name} must be >0.")
    # check for valid percentile
    if upper_percentile < 0 or upper_percentile >= 1:
        raise ValueError(
            f"upper_percentile should lie in the [0, 1) interval. {upper_percentile} provided instead!"
        )

    gm_inhibition = sts.gamma(a=inhib_a, scale=1 / inhib_b)
    gm_excitation = sts.gamma(a=excit_a, scale=1 / excit_b)

    # calculate upper bound for the evaluation
    xmax = max(gm_inhibition.ppf(upper_percentile), gm_excitation.ppf(upper_percentile))
    # equi-spaced sample covering the range
    x = np.linspace(0, xmax, ws)

    # compute difference of gammas & normalize
    gamma_diff = gm_excitation.pdf(x) - gm_inhibition.pdf(x)
    gamma_diff = gamma_diff / np.linalg.norm(gamma_diff, ord=2)

    return gamma_diff


def regress_filter(coupling_filters: NDArray, eval_basis: NDArray) -> NDArray:
    """Approximate scipy.stats.gamma based filters with basis function.

    Find the Ordinary Least Squares weights for representing the filters in terms of basis functions.

    Parameters
    ----------
    coupling_filters:
        The coupling filters. Shape ``(window_size, n_neurons_receiver, n_neurons_sender)``
    eval_basis:
        The evaluated basis function, shape ``(window_size, n_basis_funcs)``

    Returns
    -------
    weights:
        The weights for each neuron. Shape ``(n_basis_funcs, n_neurons_receiver, n_neurons_sender)``

    Raises
    ------
    ValueError
        If eval_basis is not two-dimensional.
    ValueError
        If coupling_filters is not three-dimensional.
    ValueError
        If window_size differs between eval_basis and coupling_filters.

    Examples
    --------
    .. plot::
        :include-source: True
        :caption: Least-squares approximate filter.

        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from nemos.simulation import regress_filter, difference_of_gammas
        >>> from nemos.basis import RaisedCosineLogEval
        >>> filter_duration = 100
        >>> n_basis_funcs = 20
        >>> filter_bank = difference_of_gammas(filter_duration).reshape(filter_duration, 1, 1)
        >>> _, basis = RaisedCosineLogEval(10).evaluate_on_grid(filter_duration)
        >>> weights = regress_filter(filter_bank, basis)[0, 0]
        >>> print("Weights shape:", weights.shape)
        Weights shape: (10,)
        >>> _ = plt.plot(filter_bank[:, 0, 0], label=f"True filter")
        >>> _ = plt.plot(basis.dot(weights), "--", label=f"Approx. filter")
        >>> _ = plt.legend()
        >>> _ = plt.title("True vs. Approximated Filters")
        >>> _ = plt.show()

    """
    # check shapes
    if eval_basis.ndim != 2:
        raise ValueError(
            "eval_basis must be a 2 dimensional array, "
            "shape (window_size, n_basis_funcs). "
            f"{eval_basis.ndim} dimensional array provided instead!"
        )
    if coupling_filters.ndim != 3:
        raise ValueError(
            "coupling_filters must be a 3 dimensional array, "
            "shape (window_size, n_neurons, n_neurons). "
            f"{coupling_filters.ndim} dimensional array provided instead!"
        )

    ws, n_neurons_receiver, n_neurons_sender = coupling_filters.shape

    # check that window size matches
    if eval_basis.shape[0] != ws:
        raise ValueError(
            "window_size mismatch. The window size of coupling_filters and eval_basis "
            f"does not match. coupling_filters has a window size of {ws}; "
            f"eval_basis has a window size of {eval_basis.shape[0]}."
        )

    # Reshape the coupling_filters for vectorized least-squares
    filters_reshaped = coupling_filters.reshape(ws, -1)

    # Solve the least squares problem for all filters at once
    # (vecotrizing the features)
    weights = np.linalg.lstsq(eval_basis, filters_reshaped, rcond=None)[0]

    # Reshape back to the original dimensions
    weights = np.transpose(
        weights.reshape(-1, n_neurons_receiver, n_neurons_sender), axes=(1, 2, 0)
    )

    return weights


def simulate_recurrent(
    coupling_coef: NDArray,
    feedforward_coef: NDArray,
    intercepts: NDArray,
    random_key: jax.Array,
    feedforward_input: Union[NDArray, jnp.ndarray],
    coupling_basis_matrix: Union[NDArray, jnp.ndarray],
    init_y: Union[NDArray, jnp.ndarray],
    inverse_link_function: Callable = jax.nn.softplus,
):
    """
    Simulate neural activity using the GLM as a recurrent network.

    This function projects neural activity into the future, employing the fitted
    parameters of the GLM. It is capable of simulating activity based on a combination
    of historical activity and external feedforward inputs like convolved currents, light
    intensities, etc.

    Parameters
    ----------
    coupling_coef :
        Coefficients for the coupling (recurrent connections) between neurons.
        Expected shape: (n_neurons (receiver), n_neurons (sender), n_basis_coupling).
    feedforward_coef :
        Coefficients for the feedforward inputs to each neuron.
        Expected shape: ``(n_neurons, n_basis_input)``.
    intercepts :
        Bias term for each neuron. Expected shape: ``(n_neurons,)``.
    random_key :
        jax.random.key for seeding the simulation.
    feedforward_input :
        External input matrix to the model, representing factors like convolved currents,
        light intensities, etc. When not provided, the simulation is done with coupling-only.
        Expected shape: ``(n_time_bins, n_neurons, n_basis_input)``.
    init_y :
        Initial observation (spike counts for PoissonGLM) matrix that kickstarts the simulation.
        Expected shape: ``(window_size, n_neurons)``.
    coupling_basis_matrix :
        Basis matrix for coupling, representing between-neuron couplings
        and auto-correlations. Expected shape: ``(window_size, n_basis_coupling)``.
    inverse_link_function :
        The inverse link function for the observation model.

    Returns
    -------
    simulated_activity :
        Simulated activity (spike counts for PoissonGLMs) for each neuron over time.
        Shape, ``(n_time_bins, n_neurons)``.
    firing_rates :
        Simulated rates for each neuron over time. Shape, ``(n_time_bins, n_neurons,)``.

    Raises
    ------
    ValueError
        If there's an inconsistency between the number of neurons in model parameters.
    ValueError
        If the number of neurons in input arguments doesn't match with model parameters.

    Examples
    --------
    .. plot::
        :include-source: True
        :caption: Recurrently connected GLM simulations.

        >>> import numpy as np
        >>> import jax
        >>> import matplotlib.pyplot as plt
        >>> from nemos.simulation import simulate_recurrent
        >>> np.random.seed(42)
        >>> n_neurons = 2
        >>> coupling_duration = 100
        >>> feedforward_input = np.random.normal(size=(1000, n_neurons, 1))
        >>> coupling_basis = np.random.normal(size=(coupling_duration, 10))
        >>> coupling_coef = 0.5*np.random.normal(size=(n_neurons, n_neurons, 10))
        >>> intercept = -9 * np.ones(n_neurons)
        >>> init_spikes = np.zeros((coupling_duration, n_neurons))
        >>> random_key = jax.random.key(123)
        >>> spikes, rates = simulate_recurrent(
        ...     coupling_coef=coupling_coef,
        ...     feedforward_coef=np.ones((n_neurons, 1)),
        ...     intercepts=intercept,
        ...     random_key=random_key,
        ...     feedforward_input=feedforward_input,
        ...     coupling_basis_matrix=coupling_basis,
        ...     init_y=init_spikes
        ... )
        >>> _ = plt.figure()
        >>> _ = plt.plot(rates[:, 0], label="Neuron 0 rate")
        >>> _ = plt.plot(rates[:, 1], label="Neuron 1 rate")
        >>> _ = plt.legend()
        >>> _ = plt.title("Simulated firing rates")
        >>> _ = plt.show()
    """
    if isinstance(feedforward_input, FeaturePytree):
        raise ValueError(
            "simulate_recurrent works only with arrays. "
            "FeaturePytree provided instead!"
        )
    # convert to jnp.ndarray of floats
    coupling_basis_matrix = jnp.asarray(coupling_basis_matrix, dtype=float)
    coupling_coef = jnp.asarray(coupling_coef, dtype=float)
    feedforward_coef = jnp.asarray(feedforward_coef, dtype=float)
    intercepts = jnp.asarray(intercepts, dtype=float)
    feedforward_input = jax.tree_util.tree_map(
        lambda x: jnp.asarray(x, dtype=float), feedforward_input
    )
    init_y = jnp.asarray(init_y, dtype=float)

    # check that n_neurons is consistent
    n_neurons = intercepts.shape[0]
    if (
        feedforward_input.shape[1] != n_neurons
        or feedforward_coef.shape[0] != n_neurons
        or init_y.shape[1] != n_neurons
        or coupling_coef.shape[0] != n_neurons
        or coupling_coef.shape[1] != n_neurons
    ):
        raise ValueError(
            "The number of neurons provided in the inputs is inconsistent!"
        )

    # checks the input size
    validation.check_tree_leaves_dimensionality(
        feedforward_input,
        expected_dim=3,
        err_message="`feedforward_input` must be three-dimensional, with shape "
        "(n_timebins, n_neurons, n_features) or pytree of the same shape.",
    )
    validation.check_tree_axis_consistency(
        feedforward_coef,
        feedforward_input,
        axis_1=1,
        axis_2=2,
        err_message="Inconsistent number of features. "
        f"spike basis coefficients has {jax.tree_util.tree_map(lambda p: p.shape[0], feedforward_coef)} features, "
        f"X has {jax.tree_util.tree_map(lambda x: x.shape[2], feedforward_input)} features instead!",
    )

    validation.error_invalid_entry(feedforward_input)

    # validate y
    validation.check_tree_leaves_dimensionality(
        init_y,
        expected_dim=2,
        err_message="`init_y` must be two-dimensional, with shape (n_timebins, ).",
    )
    n_basis = coupling_coef.shape[-1]
    coupling_coef = coupling_coef.reshape(n_neurons, -1)

    if coupling_basis_matrix.shape[1] * n_neurons != coupling_coef.shape[1]:
        raise ValueError(
            f"Inconsistent number of features. `coupling_basis_matrix` assumes "
            f"{coupling_basis_matrix.shape[1]} basis functions for the coupling filters, "
            f"`coupling_coef` assumes {n_basis} basis functions instead."
        )

    if init_y.shape[0] != coupling_basis_matrix.shape[0]:
        raise ValueError(
            "`init_y` and `coupling_basis_matrix`"
            " should have the same window size! "
            f"`init_y` window size: {init_y.shape[0]}, "
            f"`coupling_basis_matrix` window size: {coupling_basis_matrix.shape[0]}"
        )

    subkeys = jax.random.split(random_key, num=feedforward_input.shape[0])
    # (n_samples, n_neurons)
    feed_forward_contrib = jnp.einsum("ik,tik->ti", feedforward_coef, feedforward_input)

    def scan_fn(
        data: Tuple[jnp.ndarray, int], key: jax.Array
    ) -> Tuple[Tuple[jnp.ndarray, int], Tuple[jnp.ndarray, jnp.ndarray]]:
        """Scan over time steps and simulate activity and rates.

        This function simulates the neural activity and firing rates for each time step
        based on the previous activity, feedforward input, and model coefficients.
        """
        activity, t_sample = data

        # Convolve the neural activity with the coupling basis matrix
        # Output of shape (1, n_neuron, n_basis_coupling)
        # 1. The first dimension is time, and 1 is by construction since we are simulating 1
        #    sample
        # 2. Flatten to shape (n_neuron * n_basis_coupling, )
        # Convolution in safe mode (no vectorization)
        conv_act = convolve._tensor_convolve(
            activity,
            coupling_basis_matrix,
            batch_size_samples=feedforward_input.shape[0],
            batch_size_channels=1,
            batch_size_basis=1,
        ).reshape(
            -1,
        )

        # Extract the slice of the feedforward input for the current time step
        input_slice = jax.lax.dynamic_slice(
            feed_forward_contrib,
            (t_sample, 0),
            (1, feed_forward_contrib.shape[1]),
        ).squeeze(axis=0)

        # Predict the firing rate using the model coefficients
        # Doesn't use predict because the non-linearity needs
        # to be applied after we add the feed forward input
        firing_rate = inverse_link_function(
            coupling_coef.dot(conv_act) + input_slice + intercepts
        )

        # Simulate activity based on the predicted firing rate
        new_act = jax.random.poisson(key, firing_rate)

        # Shift of one sample the spike count window
        # for the next iteration (i.e. remove the first counts, and
        # stack the newly generated sample)
        # Increase the t_sample by one
        carry = jnp.vstack((activity[1:], new_act)), t_sample + 1
        return carry, (new_act, firing_rate)

    _, outputs = jax.lax.scan(scan_fn, (init_y, 0), subkeys)
    simulated_activity, firing_rates = outputs
    return simulated_activity, firing_rates



"""Convolution utilities."""

# required to get ArrayLike to render correctly
from __future__ import annotations

import re
import warnings
from functools import partial
from math import prod
from typing import Any, Callable, Literal, Optional

import jax
import jax.numpy as jnp
from numpy.typing import ArrayLike, NDArray

from . import type_casting, utils, validation


def _resolve_shift_default(shift, predictor_causality):
    if shift is None:
        return predictor_causality != "acausal"
    return shift


_CORR_VEC_BASIS = jax.vmap(partial(jnp.convolve, mode="valid"), (None, 1), 1)

_CORR_VEC = jax.vmap(partial(jnp.convolve, mode="valid"), (1, None), 1)
_CORR_VEC = jax.vmap(_CORR_VEC, (None, 1), 2)


def _reorganize_scan_out(out, axis):
    out = jnp.transpose(
        out, [i + 1 for i in range(axis)] + [0] + [i for i in range(axis + 1, out.ndim)]
    )
    return out.reshape(*out.shape[:axis], -1, *out.shape[axis + 2 :])


@partial(jax.jit, static_argnums=(2, 3, 4, 5, 6))
def _batch_binary_func(
    batched_array: NDArray,
    other_array: NDArray,
    binary_func: Callable[[NDArray, NDArray], NDArray],
    batch_size: int,
    axis: int = 1,
    out_axis: int = 1,
    pad_final_batch: bool = True,
):
    """
    Apply a binary function in memory-efficient batches along a specified axis.

    This function slices `batched_array` along the given `axis` into batches of size `batch_size`,
    and applies `binary_func(chunk, other_array)` to each batch. Results are concatenated
    along the same axis.

    Optionally, the final batch can be padded to avoid recompilation due to varying shapes.

    Parameters
    ----------
    batched_array :
        The array to be processed in chunks along the specified axis.
        This is typically the larger array in a binary operation.
    other_array :
        The second operand passed to `binary_func` (not batched).
        This is typically shared across batches (e.g. convolution kernel).
    binary_func :
        A function accepting two arrays (batch_chunk, other_array) and returning an array of the same shape
        as a batch result. Must be JAX-compatible and support broadcasting across batch dimensions.
    batch_size :
        Number of elements to process per batch along the batching axis.
    axis :
        The axis along which to batch `batched_array`.
    out_axis :
        The axis over which concatenating the scan output.
    pad_final_batch :
        If True, pads the final batch to `batch_size` with zeros (and slices off the padded output),
        ensuring uniform shape for all batches and preventing JAX recompilation.
        If False, the leftover portion is processed separately, which may incur compilation overhead.

    Returns
    -------
    batched_out :
        The output of applying `binary_func` to all slices of `batched_array`, concatenated along `out_axis`.

    Notes
    -----
    - This function is useful when applying memory-intensive operations (e.g., convolutions) on large arrays.
    - If `pad_final_batch` is False, JAX may compile multiple versions of `binary_func`, affecting performance.
    - If both operands must be batched, a separate batching loop should be implemented externally.

    Examples
    --------
    >>> import jax.numpy as jnp
    >>> import jax
    >>>
    >>> jax.config.update("jax_enable_x64", True)
    >>> def func(x, y):
    ...     return x + y
    >>> x, y = jax.random.normal(jax.random.PRNGKey(123), (10, 11, 12)), jnp.array(1)
    >>> result = _batch_binary_func(x, y, func, batch_size=3, axis=1)
    >>> jnp.allclose(result, x + y)
    Array(True, dtype=bool)
    """
    n_batched_features = batched_array.shape[axis]
    leftover = n_batched_features % batch_size
    n_batches = (
        (n_batched_features + batch_size - 1) // batch_size
        if pad_final_batch
        else n_batched_features // batch_size
    )

    if pad_final_batch and leftover > 0:
        pad_width = [(0, 0)] * batched_array.ndim
        pad_width[axis] = (0, batch_size - leftover)
        batched_array = jnp.pad(batched_array, pad_width, constant_values=0)

    def scan_fn(_, batch_idx):
        start = batch_idx * batch_size
        # define slices
        start_indices = [0] * axis + [start] + [0] * (batched_array.ndim - axis - 1)
        slice_sizes = (
            *batched_array.shape[:axis],
            batch_size,
            *batched_array.shape[axis + 1 :],
        )
        chunk = jax.lax.dynamic_slice(batched_array, start_indices, slice_sizes)
        conv_out = binary_func(chunk, other_array)
        return None, conv_out

    _, batched_out = jax.lax.scan(scan_fn, None, jnp.arange(n_batches))

    # Reshape
    batched_out = _reorganize_scan_out(batched_out, out_axis)

    if pad_final_batch and leftover > 0:
        batched_out = batched_out[
            (slice(None),) * out_axis
            + (slice(0, n_batched_features),)
            + (slice(None),) * (batched_out.ndim - out_axis - 1)
        ]
    elif not pad_final_batch and leftover > 0:
        # process the leftover batch manually
        slices = (
            *[slice(None)] * axis,
            slice(n_batches * batch_size, n_batched_features),
            *[slice(None)] * (batched_array.ndim - axis - 1),
        )
        leftover_out = binary_func(batched_array[slices], other_array)
        batched_out = jnp.concatenate([batched_out, leftover_out], axis=out_axis)

    return batched_out


@partial(jax.jit, static_argnums=(2, 3, 4))
def _tensor_convolve(
    array: NDArray,
    eval_basis: NDArray,
    batch_size_samples: int,
    batch_size_channels: int,
    batch_size_basis: int,
):
    """
    Memory-efficient convolution using scan over batches of input channels.

    This function first flattens the input array across dimensions other than the first one, then
    performs a vectorized convolution with the evaluation basis. The result is reshaped back to the
    original array's shape (except for the first dimension, which is adjusted based on the convolution).

    Parameters
    ----------
    array :
        The input array to convolve. It is expected to be at least 1D. The first axis is expected to be
        the sample axis, i.e. the shape of array is ``(num_samples, ...)``.
    eval_basis :
        The evaluation basis array for convolution. It should be 2D, where the first dimension
        represents the window size for convolution. Shape ``(window_size, n_basis_funcs)``.
    batch_size_samples :
        The batch size for convolution as number of samples. Must be larger than ``eval_basis.shape[1]``.
    batch_size_channels :
        Batch size over time series channels.
    batch_size_basis :
        Batch size over basis filters.

    Returns
    -------
    :
        The convolved array, reshaped to maintain the original dimensions except for the first one,
        which is adjusted based on the window size of ``eval_basis``.

    Notes
    -----
    The convolution implemented here is in mode ``"valid"``. This implies that the time axis shrinks
    ``num_samples - window_size + 1``, where num_samples is the first size of the first axis of ``array``
    and ``window_size`` is the size of the first axis in ``eval_basis``.
    """
    n_samples, *feat_shape = array.shape

    array = array.reshape(n_samples, -1)
    window_size, n_basis = eval_basis.shape
    n_features = array.shape[1]

    step = batch_size_samples - window_size + 1
    n_samples_padded = (
        ((n_samples - window_size + 1 + step - 1) // step) * step + window_size - 1
    )
    n_batches = (n_samples_padded - window_size + 1) // step

    pad_len = n_samples_padded - n_samples
    array = jnp.pad(array, ((0, pad_len), (0, 0)), constant_values=0.0)

    def conv_batch(concat_conv, batch_idx):
        start_chunk = batch_idx * step
        chunk = jax.lax.dynamic_slice(
            array, (start_chunk, 0), (batch_size_samples, n_features)
        )
        concat_conv = jax.lax.dynamic_update_slice(
            concat_conv,
            _batch_convolve_over_channels(
                chunk, eval_basis, batch_size_channels, batch_size_basis
            ),
            (start_chunk, 0, 0),
        )
        return concat_conv, None

    # initialize output of valid convolution
    conv_output = jnp.full(
        (n_samples_padded - window_size + 1, n_features, n_basis), jnp.nan
    )
    conv_output, _ = jax.lax.scan(conv_batch, conv_output, jnp.arange(n_batches))
    conv_output = conv_output[: n_samples - window_size + 1]
    return conv_output.reshape(n_samples - window_size + 1, *feat_shape, n_basis)


def _batch_convolve_over_channels(
    array: NDArray, eval_basis: NDArray, batch_size_channels: int, batch_size_basis: int
):
    def conv_over_basis(array_chunk, basis_matrix):
        return _batch_over_basis_convolve(array_chunk, basis_matrix, batch_size_basis)

    return _batch_binary_func(
        array, eval_basis, conv_over_basis, batch_size_channels, axis=1
    )


def _batch_over_basis_convolve(
    array: NDArray, eval_basis: NDArray, batch_size_basis: int
):
    out = _batch_binary_func(
        eval_basis,
        array,
        _CORR_VEC,
        batch_size_basis,
        axis=1,
        out_axis=1,
    )
    # move basis axis.
    return out.transpose(0, 2, 1)


def _shift_time_axis_and_convolve(
    array: NDArray,
    eval_basis: NDArray,
    axis: int,
    batch_size_samples: int,
    batch_size_channels: int,
    batch_size_basis: int,
):
    """
    Shifts the specified axis to the first position, applies convolution, and then reverses the shift.

    This applies a convolution along a specific axis of a multi-dimensional array. The process involves
    three steps: shifting the axis, convolving, and then reversing the shift to maintain the original
    axis order.

    Parameters
    ----------
    array :
        The input array for convolution.
    eval_basis :
        The evaluation basis array for convolution, should be 2D.
    axis :
        The axis along which the convolution is applied. This axis is temporarily shifted
        to the first position for the convolution operation.
    batch_size_samples :
        Size of the batches in samples.
    batch_size_channels :
        Size of the batches in number of input channels.
    batch_size_basis :
        Size of the batches in number of basis filters.

    Returns
    -------
    conv :
        The array after applying the convolution along the specified axis, with the original
        axis order restored.

    Notes
    -----
    This function supports arrays of any dimensionality greater or equal than 1.
    """
    # convert axis
    axis = axis if axis >= 0 else array.ndim + axis
    # move time axis to first
    new_axis = (jnp.arange(array.ndim) + axis) % array.ndim
    array = jnp.transpose(array, new_axis)

    # convolve
    conv = _tensor_convolve(
        array, eval_basis, batch_size_samples, batch_size_channels, batch_size_basis
    )

    # reverse transposition
    new_axis = (*((jnp.arange(array.ndim) - axis) % array.ndim), array.ndim)
    conv = jnp.transpose(conv, new_axis)
    return conv


def _list_epochs(tsd: Any):
    """
    List epochs from a time series with data object, supporting 'pynapple' Tsd formats.

    If the input is recognized as a 'pynapple' TSD, it extracts epochs based on the TSD's
    time support. Otherwise, it returns the input as is, assuming it's a single epoch.

    Parameters
    ----------
    tsd :
        The time series data, potentially in Pynapple TSD format or a generic data structure.

    Returns
    -------
    :
        A list of epochs extracted from the TSD. If the input is not a Pynapple TSD, returns
        a list containing the input itself.
    """
    if type_casting.is_pynapple_tsd(tsd):
        return [tsd.get(s, e) for s, e in tsd.time_support.values]
    return [tsd]


def _convolve_pad_and_shift(
    basis_matrix: ArrayLike,
    time_series: Any,
    batch_size_samples: int,
    batch_size_channels: Any,
    batch_size_basis: int,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    axis: int = 0,
    shift: Optional[bool] = None,
):
    """
    Create predictor by convolving basis_matrix with time_series.

    To create the convolutional predictor, three steps are taken, each of
    calls a single function. See their docstrings for more details about
    each step. This function **preserves** the number of samples by
    NaN-padding appropriately.

    - Convolve `basis_matrix` with `time_series` (function: `convolve_1d_trials`)
    - Pad output with `basis_matrix.shape[0]-1` NaNs, with location based on
      causality of intended predictor (function: `nan_pad`).
    - (Optional) Shift predictor based on causality (function: `shift_time_series`)

    Parameters
    ----------
    basis_matrix :
        The basis matrix with which to convolve the trials. Shape
        `(window_size, n_basis_funcs)`.
    time_series :
        The time series to convolve with the basis matrix. This variable should
        be a pytree with arrays of at least one-dimension as leaves.
    batch_size_samples:
        Size of the batches in samples.
    batch_size_channels :
        Pytree of batch sizes. The number of batched channels for the convolution.
    batch_size_basis :
        Number of batched basis filters for the convolution.
    predictor_causality:
        Causality of this predictor, which determines where padded values are
        added and how the predictor is shifted.
    axis:
        Axis containing samples.
    shift :
        Whether to shift predictor based on causality (only valid if
        `predictor_causality != 'acausal'`). Default is True for `causal` and
        `anti-causal`, False for `acausal`.

    Returns
    -------
    predictor :
        Predictor of with same shape and structure as `time_series`
    """

    # apply convolution
    def conv(x, bs, bc):
        return _shift_time_axis_and_convolve(
            x,
            basis_matrix,
            axis=axis,
            batch_size_samples=bs,
            batch_size_channels=bc,
            batch_size_basis=batch_size_basis,
        )

    predictor = jax.tree_util.tree_map(
        conv, time_series, batch_size_samples, batch_size_channels
    )

    with warnings.catch_warnings(record=True) as warns:
        warnings.simplefilter("always")
        predictor = utils.nan_pad(
            predictor, basis_matrix.shape[0] - 1, predictor_causality, axis=axis
        )

    for w in warns:
        if re.match("^With acausal filter", str(w.message)):
            warnings.warn(
                message="With `acausal` filter, `basis_matrix.shape[0] "
                "should probably be odd, so that we can place an equal number of NaNs on "
                "either side of input.",
                category=UserWarning,
            )
        else:
            warnings.warn(w.message, w.category)

    if shift:
        predictor = utils.shift_time_series(predictor, predictor_causality, axis=axis)
    return predictor


def create_convolutional_predictor(
    basis_matrix: ArrayLike,
    time_series: Any,
    predictor_causality: Literal["causal", "acausal", "anti-causal"] = "causal",
    shift: Optional[bool] = None,
    axis: int = 0,
    batch_size_samples: Optional[int] = None,
    batch_size_channels: Optional[int] = None,
    batch_size_basis: Optional[int] = None,
):
    """
    Create a convolutional predictor by convolving a basis matrix with a time series.

    To create the convolutional predictor, three steps are taken, convolve, pad and shift.

    - Convolve `basis_matrix` with `time_series` (function: `_convolve_1d_trials`)
    - Pad output with `basis_matrix.shape[0]-1` NaNs, with location based on
      causality of intended predictor (function: `nemos.utils.nan_pad`).
    - (Optional) Shift predictor based on causality (function: `nemos.utils.shift_time_series`)

    The function is designed to handle both single arrays and pynapple time series with data
     (across multiple epochs), as well as their combinations. For pynapple time series, it
    treats each epoch separately, applies the convolution process, and then reassembles the time series.

    Parameters
    ----------
    basis_matrix :
        A 2D array representing the basis matrix to convolve with the time series.
        The first dimension should represent the window size for the convolution.
    time_series :
        The time series data to convolve with the basis matrix. Can be single arrays, pytree of arrays,
        pynapple time series, pytree of pynapple time series or a mix.
        In case of Pynapple time series data, each epoch will be convolved separately.
    predictor_causality :
        The causality of the predictor, determining how the padding and shifting
        should be applied to the convolution result.
        - 'causal': Pads and/or shifts the result to be causal with respect to the input.
        - 'acausal': Applies padding equally on both sides without shifting.
        - 'anti-causal': Pads and/or shifts the result to be anti-causal with respect to the input.
    shift :
        Determines whether to shift the convolution result based on the causality.
        If None, it defaults to True for 'causal' and 'anti-causal' and to False for 'acausal'.
    axis :
        The axis along which the convolution is applied.
    batch_size_samples :
        Batch size for the convolution in terms of number of sample points.
        If this parameter is set, the convolution will be applied sequentially
        over a fixed-length chunks of the time_series.
    batch_size_channels :
        Batch size for the convolution in terms of number of input channels.
        If this parameter is set, the convolution will be vectorized over batches
        of input channels. Default vectorizes over all channels.
    batch_size_basis :
        Batch size for the convolution in terms of number of basis kernels.
        If this parameter is set, the convolution will be vectorized over batches
        of kernels. Default vectorizes over all basis kernels.

    Returns
    -------
    :
        The convolutional predictor, structured similarly to the input `time_series`, appropriately
        padded and shifted according to the specified causality.

    Raises
    ------
    ValueError:
        If `basis_matrix` is not a 2-dimensional array or has a singleton first dimension.
    ValueError:
        If `time_series` does not contain arrays of at least one dimension or contains
        arrays with a dimensionality less than `axis`.
    ValueError:
        If any array within `time_series` or `basis_matrix` is empty.
    ValueError:
        If the number of elements along the convolution axis in any array within `time_series`
        is less than the window size of the `basis_matrix`.
    ValueError:
        If shifting is attempted with 'acausal' causality.
    ValueError:
        Raised if any explicitly provided batch sizes—``batch_size_samples``,
        ``batch_size_channels``, or ``batch_size_basis``—are not positive integers.
        A value of ``None`` is allowed and treated as unspecified.

    """
    # apply checks
    validation._check_basis_matrix_shape(basis_matrix)
    validation._check_non_empty_inputs(time_series, basis_matrix)
    validation._check_time_series_ndim(time_series, axis)
    shift = _resolve_shift_default(shift, predictor_causality)
    validation._check_shift_causality_consistency(shift, predictor_causality)
    validation._check_batch_size(batch_size_samples, "batch_size_samples")
    validation._check_batch_size(batch_size_channels, "batch_size_channels")
    validation._check_batch_size(batch_size_basis, "batch_size_basis")

    # flatten and grab tree struct
    time_series, struct = jax.tree_util.tree_flatten(time_series)

    # run a single conv if batches provided, otherwise make sure batch size is smaller than
    # array shape.
    if batch_size_samples is None:
        batch_size_samples = jax.tree_util.tree_map(
            lambda x: x.shape[axis], time_series
        )
    else:
        batch_size_samples = jax.tree_util.tree_map(
            lambda x: min([x.shape[axis], batch_size_samples]), time_series
        )

    if batch_size_channels is None:
        batch_size_channels = jax.tree_util.tree_map(
            lambda x: prod(x.shape[:axis] + x.shape[axis + 1 :]), time_series
        )
    else:
        batch_size_channels = jax.tree_util.tree_map(
            lambda x: min(
                [prod(x.shape[:axis] + x.shape[axis + 1 :]), batch_size_channels]
            ),
            time_series,
        )

    # basis is always a 2D array, no tree-map needed for the same logic
    if batch_size_basis is None:
        batch_size_basis = basis_matrix.shape[1]
    else:
        batch_size_basis = min([batch_size_basis, basis_matrix.shape[1]])

    # find pynapple
    is_nap = list(type_casting.is_pynapple_tsd(x) for x in time_series)

    # retrieve time info
    time_info = [
        type_casting.get_time_info(ts) if is_nap[i] else None
        for i, ts in enumerate(time_series)
    ]

    # split epochs (adds one layer to pytree)
    # if pynapple one batch size per epoch to match tree-struct
    def add_tree_level(x, y):
        return [x] * len(y.time_support) if type_casting.is_pynapple_tsd(y) else [x]

    batch_size_channels = jax.tree_util.tree_map(
        add_tree_level,
        batch_size_channels,
        time_series,
    )
    batch_size_samples = jax.tree_util.tree_map(
        add_tree_level,
        batch_size_samples,
        time_series,
    )
    time_series = jax.tree_util.tree_map(_list_epochs, time_series)

    # check trial size (after splitting)
    validation._check_trials_longer_than_time_window(
        time_series, basis_matrix.shape[0], axis
    )
    validation._check_batch_size_larger_than_convolution_window(
        batch_size=batch_size_samples, window_size=basis_matrix.shape[0]
    )

    # convert to array
    time_series = jax.tree_util.tree_map(jnp.asarray, time_series)

    # convolve
    conv = _convolve_pad_and_shift(
        basis_matrix,
        time_series,
        batch_size_samples=batch_size_samples,
        batch_size_channels=batch_size_channels,
        batch_size_basis=batch_size_basis,
        predictor_causality=predictor_causality,
        axis=axis,
        shift=shift,
    )

    #  concatenate back
    flat_stack = [jnp.concatenate(x, axis=axis) for x in conv]

    # re-attach time axis
    flat_stack = [
        type_casting.cast_to_pynapple(x, *time_info[i]) if is_nap[i] else x
        for i, x in enumerate(flat_stack)
    ]

    # recreate tree
    return jax.tree_util.tree_unflatten(struct, flat_stack)
